function httpGet(url, token) {
  return new Promise((resolve, reject) => {
    var xhr = new XMLHttpRequest();
    xhr.open("GET", url, true);

    xhr.setRequestHeader("Authorization", "Bearer " + token);

    xhr.onreadystatechange = function() {
      if(this.readyState == XMLHttpRequest.DONE) {
        if (this.status == 200) {
          resolve(this.responseText);
        } else {
          reject(this.responseText);
        }
      }
    }

    xhr.send();
  });
}

class RequestError extends Error {
  constructor(response, status, statusText) {
    super(response);
    this.response = response;
    this.status = status;
    this.statusText = statusText;
  }
}

/**
 * Make a XHR Request to for JSON data at the specified url.
 * body is an optional string or object, if provided the method will
 * be POST, GET otherwise. If succeded, the response will be parsed as JSON.
 */
function jsonRequest(url, body) {
  return new Promise((resolve, reject) => {
    var xhr = new XMLHttpRequest();

    if (body) {
      xhr.open("POST", url, true);
    } else {
      xhr.open("GET", url, true);
    }
    
    xhr.setRequestHeader("Content-Type", "application/json");

    xhr.onreadystatechange = function() {
      if(this.readyState == XMLHttpRequest.DONE) {
        if (this.status == 200) {
          try {
            const jsonResult = JSON.parse(this.responseText);
            resolve(jsonResult);
          } catch (err) {
            reject(new RequestError("Failed to parse " + url + " response as JSON. " + err.message, this.status, this.statusText));
          }
        } else {
          reject(new RequestError(this.responseText || "Request error. Status code: " + this.status, this.status, this.statusText));
        }
      }
    }

    if (body) {
      const textBody = typeof body === "string" ? body : JSON.stringify(body);
      xhr.send(textBody);
    } else {
      xhr.send();
    }
  });
}

/**
 * Make a XHR Request to the given URL. 180 sec retry time.
 * body is an optional string or object, if provided the method will
 * be POST, GET otherwise.
 */
 function xhrPolling(url, method, body, timeout) {
  return new Promise((resolve, reject) => {
		timeoutTime = !!timeout && timeout !== 0 ? timeout : 180000;
		if (timeoutTime <= 0) {
      reject("Request has timed out.");
      return;
    }
    const dt = new Date();
    var xhr = new XMLHttpRequest();

    xhr.onreadystatechange = function() {
      if(this.readyState == XMLHttpRequest.DONE) {
					if (xhr.status === 200) {
						if (!xhr.response) {
              const remainingTime = timeoutTime - (Date.now() - dt.getTime());
              resolve(xhrPolling(url, method, body, remainingTime))
						}
						try {
							resolve(xhr.response);
						} catch (err) {
							reject(new RequestError("Failed to parse " + url + " response as JSON. " + err.message + "xhr.response", this.status, this.statusText));
						}
					} else {
						reject(new RequestError("Request error. Status code: " + this.status, this.statusText));
					}
      }
    }
		xhr.ontimeout = () => {
      const remainingTime = timeoutTime - (Date.now() - dt.getTime());
      resolve(xhrPolling(url, method, body, remainingTime));
    };
		xhr.onerror = (err) => {
			reject(err)
    }
    xhr.open(method, url, true);
    xhr.setRequestHeader('Content-Type', 'application/x-www-form-urlencoded');
    xhr.responseType = 'json';

		if (body) {
      xhr.send(body);
    } else {
      xhr.send();
    }
  });
}

function getUserWorkspaces(rootUrl, username, token) {
  return httpGet(rootUrl + "/users/" + username + "/workspaces", token)
    .then((res) => {
      const apiRes = JSON.parse(res);
      if (apiRes.status === 0) {
        return apiRes.data.workspaces;
      } else {
        throw new Error("Worskpaces API status code: " + apiRes.status);
      }
    });
}

function getAllShares(rootUrl, username, token) {
  return new Promise((resolve, reject) => {
    var allShares = [];
    httpGet(rootUrl + "/users/" + username + "/shares", token)
      .then((r) => {
        var personalShares = JSON.parse(r).data.shares;
        allShares = personalShares;

        getUserWorkspaces(rootUrl, username, token)
          .then((workspaces) => {
            var promises = [];
            workspaces.forEach((ws) => {
              var p = httpGet(rootUrl + '/workspaces/' + ws.workspaceId + '/shares', token);
              promises.push(p);
            });

            Promise.all(promises)
              .then((result) => {
                result.forEach((sr) => {
                  allShares = allShares.concat(JSON.parse(sr).data.shares);
                });

                resolve({
                  shares: allShares,
                  workspaces: workspaces
                });
              })
              .catch((err) => {
                reject(err);
              });
          })
          .catch((err) => {
            reject(err);
          });
      })
      .catch((err) => {
        reject(err);
      });
  });
}

function getDeviceIds(url) {
  return new Promise((resolve, reject) => {
    var xhr = new XMLHttpRequest();

    xhr.open("POST", url, true);
    xhr.setRequestHeader("Content-Type", "application/json");

    xhr.onreadystatechange = function() {
      if(this.readyState == XMLHttpRequest.DONE) {
        if (this.status == 200) {
          resolve(this.responseText);
        } else {
          reject(this.responseText);
        }
      }
    }

    var data = {
      "query": "{ devices { id, name, category, size, supportsCustomSize } }"
    };
    xhr.send(JSON.stringify(data));
  });
}

function getPortsIds(url, token) {
  return new Promise((resolve, reject) => {
    var xhr = new XMLHttpRequest();

    xhr.open("GET", url, true);
    xhr.setRequestHeader("Content-Type", "application/json");
    xhr.setRequestHeader("Authorization", "Bearer " + token);

    xhr.onreadystatechange = function() {
      if(this.readyState == XMLHttpRequest.DONE) {
        if (this.status == 200) {
          resolve(this.responseText);
        } else {
          reject(this.responseText);
        }
      }
    }

    var data = {
      "query": "{ [ports] { id, name, width, height, order, type} }"
    };
    xhr.send(JSON.stringify(data));
  });
}

function getUsabilityStudyURL(that) {
  return new Promise((resolve, reject) => {
    var userInfo = JSON.parse(localStorage.getItem("userInfo"));

    var xhr = new XMLHttpRequest();
    xhr.open("POST", that.app.apiUrl + "/studies/", true);
    xhr.setRequestHeader("Authorization", "Bearer " + userInfo.data.token);
    xhr.setRequestHeader("Content-Type", "application/json");

    xhr.onreadystatechange = function() {
      if(this.readyState == XMLHttpRequest.DONE) {
        if (this.status == 200) {
          var result = JSON.parse(this.responseText);

          var resultUrl = "";

          var workspaceId = result.data.study.workspaceId;
          if (workspaceId == null) {
            resultUrl = that.app.rootUrl + "/home/usability/" + result.data.study.studyId;
          } else {
            resultUrl = that.app.rootUrl + "/workspace/" + workspaceId + "/usability/" + result.data.study.studyId;
          }
          resolve(resultUrl);
        } else {
          reject(this.responseText);
        }
      }
    }

    var shareId = that.app.publishedShareId;
    var title = that.app.publishedShareTitle;

    var body = JSON.stringify({ shareId, title });
    xhr.send(body);
  });
}

/** Creates an app in the AppBuilder */
function create(url, sketchFile, params, token) {

  return new Promise((resolve, reject) => {
    const fileName = sketchFile.filePath.split("/").slice(-1)[0];
    var viewportSizes = params.project.viewPort.split(" x ")
    var viewport = {
      width: +viewportSizes[0],
      height: +viewportSizes[1]
    }

    var formData = new FormData();
    formData.append('workspaceId', params.workspaceId);
    formData.append('viewport', JSON.stringify(viewport));
    formData.append("sketch", sketchFile.file, params.project.name);

    
    var xhr = new XMLHttpRequest();
    xhr.open("POST", url, true);

    if (token != null && token != "") {
      xhr.setRequestHeader("Authorization", "Bearer " + token);
    }

    xhr.onreadystatechange = function() {
      if(this.readyState == XMLHttpRequest.DONE) {
        if (this.status == 200) {
          resolve(this.responseText);
        } else {
          reject(this.status + ": " + this.statusText);
        }
    }
  }

    xhr.send(formData);
  });
}

/** Publish the prototype
 * @param url target API URL
 * @param params JSON object with necessary publish parameters
 * @param assets Array<{ blob, info }> containing the blob and the asset info
 * @param token authentication JWT to be sent in the authorization header
 */
 function publish(url, params, assets, token) {

  return new Promise((resolve, reject) => {
    var isNewProject = params.project.id == null || params.project.id == "";

    var endUrl = isNewProject ?
        url + "/ibp/project" :
        url + "/ibp/project/" + params.project.id;

    const fileName = sketchData.sketchFile.filePath.split("/").slice(-1)[0];
    var formData = new FormData();
    formData.append("data", JSON.stringify(params));

    // Append SketchFile
    formData.append("sketchFile", sketchData.sketchFile.file, fileName);

    // Append rendered Screens
    sketchData.screens.forEach((s) => {
      formData.append(s.id, s.image, s.name + ".png");
    });

    // Append Assets (asset info is already loaded in params)
    for (const asset of assets) {
      const blob = new Blob([asset.blob], { type: "application/octet-stream" });
      formData.append("asset", blob, asset.info.name);
    }

    var xhr = new XMLHttpRequest();
    xhr.open(isNewProject ? "POST" : "PUT", endUrl, true);

    if (token != null && token != "") {
      xhr.setRequestHeader("Authorization", "Bearer " + token);
    }

    xhr.onreadystatechange = function() {
      if(this.readyState == XMLHttpRequest.DONE) {
        if (this.status == 200) {
          resolve(this.responseText);
        } else {
          reject(this.responseText);
        }
      }
    }

    xhr.send(formData);
  });
}

function loadThumbs() {
  const promises = sketchData.screens.map(s => loadFile(s.fileName));
  return Promise.all(promises);
}

function loadSketchFile() {
  return new Promise((resolve, reject) => {
    if (!sketchData.sketchFile.filePath || sketchData.sketchFile.filePath == "null") {
      reject(new Error("No sketch filePath"));
      return;
    }
    
    return loadFile(sketchData.sketchFile.filePath).then(resolve, reject);
  });
}

function loadFile(filePath) {
  const fileUrl =
    filePath.startsWith("file:")
      ? new URL(filePath)
      : filePath.startsWith("/")
        ? new URL("file://" + filePath)
        : new URL(filePath, window.location);
  
  // Include a cache buster parameter, because WebView seems to ignore Cache-Control for file urls
  fileUrl.searchParams.append("cacheBuster", new Date().getTime());

  return fetch(fileUrl, {
    headers: {
      "Cache-Control": "no-cache, no-store, max-age=0"
    }
  }).then(res => {
    if (res.status === 200 || res.status === 0 && fileUrl.protocol === "file:") {
      return res.blob();
    } else {
      throw new Error("Couldn't load file " + filePath + ". Status: " + res.status + " " + res.statusText);
    }
  });
}
